package be.iminds.ilabt.jfed.experimenter_gui.preferences.subsections;

import be.iminds.ilabt.jfed.experimenter_gui.preferences.AbstractPreferencesSubPane;
import be.iminds.ilabt.jfed.experimenter_gui.util.ui.TextFieldWithStatus;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.user.GeniUser;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.PuTTYFilesKeyInfo;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.SshKeyInfoFactory;
import be.iminds.ilabt.jfed.lowlevel.ssh_key_info.putty.PuTTYPrivateKeyFile;
import be.iminds.ilabt.jfed.preferences.CorePreferenceKey;
import be.iminds.ilabt.jfed.gui_preferences.HLPreferenceKey;
import be.iminds.ilabt.jfed.gui_preferences.JFedHLPreferences;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.PageantHelper;
import be.iminds.ilabt.jfed.ssh_terminal_tool.putty.SshKeyInfoUnlocker;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.ui.javafx.util.InformationLabel;
import be.iminds.ilabt.jfed.util.common.IOUtils;
import be.iminds.ilabt.jfed.util.common.OSDetector;
import be.iminds.ilabt.jfed.util.library.KeyUtil;
import javafx.beans.binding.BooleanBinding;
import javafx.beans.property.BooleanProperty;
import javafx.beans.property.SimpleBooleanProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.beans.property.SimpleStringProperty;
import javafx.fxml.FXML;
import javafx.scene.Node;
import javafx.scene.control.Button;
import javafx.scene.control.CheckBox;
import javafx.scene.control.Label;
import javafx.scene.control.TextArea;
import javafx.stage.FileChooser;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.awt.event.ActionEvent;
import java.io.File;
import java.io.IOException;
import java.security.interfaces.RSAPrivateKey;
import java.security.interfaces.RSAPublicKey;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

public class SshAuthenticationPane extends AbstractPreferencesSubPane {
    private static final Logger LOG = LoggerFactory.getLogger(SshAuthenticationPane.class);

    private final OSDetector.OS os;

    private enum SshAuthState {
        PRIV_UNSPECIFIED,
        PRIV_BAD,
        PRIV_PUTTY,
        PRIV_OPENSSH_PUB_INPRIV,
        PRIV_OPENSSH_PUB_EXTERNAL_AUTO,
        PRIV_OPENSSH_PUB_EXTERNAL_MANUAL,
    }
    private final SimpleObjectProperty<SshAuthState> state = new SimpleObjectProperty<>(SshAuthState.PRIV_UNSPECIFIED);
    private final SimpleStringProperty privateKeyFilename = new SimpleStringProperty();
    private final SimpleStringProperty publicKeyFilename = new SimpleStringProperty();
    private final SimpleStringProperty privateKeyContent = new SimpleStringProperty();
    private final SimpleStringProperty publicKeyContent = new SimpleStringProperty();

    private final BooleanProperty badPrivateKey = new SimpleBooleanProperty(false);
    private final BooleanProperty badPublicKey = new SimpleBooleanProperty(false);
    private final BooleanProperty requirePublicKeyfile = new SimpleBooleanProperty(false);

    private final BooleanProperty useCustomKey = new SimpleBooleanProperty(false);
    private final BooleanProperty useCertKey = new SimpleBooleanProperty(false);

    @FXML protected Node root;
    @FXML protected CheckBox useCertificateCheckBox;
    @FXML protected CheckBox useCustomKeyPairCheckBox;

    @FXML protected TextFieldWithStatus privKeyFileTextField;
    @FXML protected Button privKeyBrowseButton;
    @FXML protected Label warnNoLoginKeyLabel;

    @FXML protected TextFieldWithStatus pubKeyFileTextField;
    @FXML protected Button pubKeyBrowseButton;
    @FXML protected Label pubKeyLabel;
    @FXML protected InformationLabel pubKeyInPrivKeyInfo;
    @FXML protected Label publicKeyContentLabel;
    @FXML protected TextArea pubKeyContentTextArea;

    protected final GeniUserProvider geniUserProvider;
    protected final JFedHLPreferences jFedPreferences;

    @FXML protected Label warnMultiKeyLabel;

    private final PageantHelper pageantHelper;
    private final SshKeyInfoUnlocker sshKeyInfoUnlocker;

    @Inject
    SshAuthenticationPane(GeniUserProvider geniUserProvider, JFedHLPreferences jFedPreferences,
                          PageantHelper pageantHelper, SshKeyInfoUnlocker sshKeyInfoUnlocker) {
        super("SSH Authentication", false);
        this.pageantHelper = pageantHelper;
        this.sshKeyInfoUnlocker = sshKeyInfoUnlocker;
        this.geniUserProvider = geniUserProvider;
        this.jFedPreferences = jFedPreferences;
        this.os = OSDetector.findOS();

        load();
    }

    private void load() {
        useCertKey.set(jFedPreferences.getBoolean(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY));
        useCustomKey.set(jFedPreferences.getBoolean(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY));

        if (jFedPreferences.getBoolean(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY) == Boolean.TRUE) {
            File privPuttyKeyFile = jFedPreferences.getFile(CorePreferenceKey.PREF_SSH_PUTTY_KEY_FILE);
            File privOpenSshKeyFile = jFedPreferences.getFile(CorePreferenceKey.PREF_SSH_PRIVATE_KEY_FILE);
            if (privPuttyKeyFile != null) {
                loadPrivKey(privPuttyKeyFile);
            } else if (privOpenSshKeyFile != null) {
                loadPrivKey(privOpenSshKeyFile);
            } else {
                loadPrivKey(null);
            }
        } else {
            loadPrivKey(null);
        }

        File publicKeyManual = jFedPreferences.getFile(CorePreferenceKey.PREF_SSH_PUBLIC_KEY_FILE);
        loadPubKey(publicKeyManual);
    }

    private void loadPrivKey(@Nullable File newPrivateKeyFile) {
        boolean setBad;
        if (newPrivateKeyFile != null) {
            String content = IOUtils.fileToStringOrNull(newPrivateKeyFile);
            privateKeyFilename.set(newPrivateKeyFile.getPath());
            privateKeyContent.set(content);
            try {
                if (content != null && testPuttyPrivateKeyFile(newPrivateKeyFile, content)) {
                    this.state.set(SshAuthState.PRIV_PUTTY);
                    loadPubKey(null);
                    setBad = false;
                } else if (content != null && testOpenSshPrivateKeyFile(newPrivateKeyFile, content)) {
                    this.state.set(SshAuthState.PRIV_OPENSSH_PUB_INPRIV); //tmp
                    loadPubKey(publicKeyFilename.get() == null ? null : new File(publicKeyFilename.get()));
                    setBad = false;
                } else {
                    this.state.set(SshAuthState.PRIV_BAD);
                    setBad = true;
                }
            } catch (IOException e) {
                this.state.set(SshAuthState.PRIV_BAD);
                setBad = true;
                privateKeyFilename.set(null);
            }
        } else {
            setBad = true;
            this.state.set(SshAuthState.PRIV_UNSPECIFIED);
            privateKeyFilename.set(null);
        }

        badPrivateKey.set(setBad);
        if (setBad) {
            privateKeyContent.set(null);
            publicKeyContent.set(null);
            publicKeyFilename.set(null);
        }

        LOG.debug("Private key loaded. state={} privateKeyFilename={} badPrivateKey={}",
                this.state.get(), privateKeyFilename.get(), badPrivateKey.get());
    }

    private void loadPubKey(@Nullable File newPublicKeyFile) {
        File privKey = privateKeyFilename.get() == null ? null: new File(privateKeyFilename.get());

        if (privKey == null || privateKeyContent.get() == null) {
            publicKeyFilename.set(null);
            publicKeyContent.set(null);
            badPublicKey.set(true);
            return;
        }

        if (state.get() != SshAuthState.PRIV_PUTTY) {
            if (testOpenSSHPrivateKeyIncludesPublicKey(privKey, privateKeyContent.get())) {
                this.state.set(SshAuthState.PRIV_OPENSSH_PUB_INPRIV);
                try {
                    this.publicKeyContent.set(getPublicKeyFromOpenSSHPrivKey(privKey, privateKeyContent.get()));
                    badPublicKey.set(false);
                } catch (KeyUtil.PEMDecodingException e) {
                    LOG.error("Unexpected PEM decoding exception while extracting public key from private key file", e);
                    badPrivateKey.set(true);
                    badPublicKey.set(true);
                    requirePublicKeyfile.set(true);
                    this.publicKeyContent.set("<bug extracting public key from private key>");
                }
            } else if (privateKeyFilename.get() != null) {
                File autoPubkeyFile = new File(privateKeyFilename.get() + ".pub");
                if (autoPubkeyFile.isFile() && autoPubkeyFile.exists()) {
                    this.publicKeyFilename.set(autoPubkeyFile.getPath());
                    try {
                        this.publicKeyContent.set(IOUtils.fileToString(autoPubkeyFile));
                        badPublicKey.set(false);
                        this.state.set(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_AUTO);
                        requirePublicKeyfile.set(false);
                    } catch (IOException e) {
                        publicKeyContent.set(null);
                        badPublicKey.set(true);
                        this.state.set(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_MANUAL);
                        requirePublicKeyfile.set(true);
                    }
                } else {
                    this.state.set(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_MANUAL);
                    requirePublicKeyfile.set(true);
                }
            } else {
                this.state.set(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_MANUAL);
                requirePublicKeyfile.set(true);
            }

            if (requirePublicKeyfile.get()) {
                if (newPublicKeyFile != null) {
                    String publicKeyManualContent = IOUtils.fileToStringOrNull(newPublicKeyFile);
                    publicKeyFilename.set(newPublicKeyFile.getPath());
                    publicKeyContent.set(publicKeyManualContent);
                    badPublicKey.set(publicKeyManualContent == null);
                } else {
                    publicKeyFilename.set(null);
                    publicKeyContent.set(null);
                    badPublicKey.set(true);
                }
            }
        } else {
            requirePublicKeyfile.set(false);
            publicKeyFilename.set(null);
            publicKeyContent.set(getPublicKeyFromPuTTYFile(privKey, privateKeyContent.get()));
            badPublicKey.set(false);
        }

        LOG.debug("Public key loaded. requirePublicKeyfile={} publicKeyFilename={} badPublicKey={} publicKeyContent={}",
                this.requirePublicKeyfile.get(), publicKeyFilename.get(), badPublicKey.get(), publicKeyContent.get());
    }

    private void updateStatus() {
        if (badPrivateKey.get()) {
            privKeyFileTextField.setStatus(TextFieldWithStatus.Status.ERROR);
        } else {
            privKeyFileTextField.setStatus(TextFieldWithStatus.Status.OK);
        }
        if (badPublicKey.get()) {
            pubKeyFileTextField.setStatus(TextFieldWithStatus.Status.ERROR);
        } else {
            pubKeyFileTextField.setStatus(TextFieldWithStatus.Status.OK);
        }
    }

    private void bindManagedToVisible(Node n) {
        n.managedProperty().bind(n.visibleProperty());
    }

    @FXML
    protected void initialize() {
        useCustomKeyPairCheckBox.selectedProperty().bindBidirectional(useCustomKey);
        useCertificateCheckBox.selectedProperty().bindBidirectional(useCertKey);

        privKeyFileTextField.disableProperty().bind(useCustomKey.not());
        privKeyBrowseButton.disableProperty().bind(useCustomKey.not());

        privKeyFileTextField.textProperty().bind(privateKeyFilename);
        pubKeyFileTextField.textProperty().bind(publicKeyFilename);
        pubKeyContentTextArea.textProperty().bind(publicKeyContent);

        //filenames are not directly editable anymore.
//        privKeyFileTextField.focusedProperty().addListener((observableValue, oldIsFocused, newIsFocused) -> {
//            if (newIsFocused) {
//                privKeyFileTextField.setStatus(TextFieldWithStatus.Status.NONE);
//            } else {
//                loadPrivKey(new File(privKeyFileTextField.getText()));
//                updateStatus();
//            }
//        });
//        pubKeyFileTextField.focusedProperty().addListener((observableValue, oldIsFocused, newIsFocused) -> {
//            if (newIsFocused) {
//                pubKeyFileTextField.setStatus(TextFieldWithStatus.Status.NONE);
//            } else {
//                loadPubKey(new File(pubKeyFileTextField.getText()));
//                updateStatus();
//            }
//        });
        badPrivateKey.addListener((observable, oldValue, newValue) -> updateStatus());
        badPublicKey.addListener((observable, oldValue, newValue) -> updateStatus());

        warnNoLoginKeyLabel.visibleProperty().bind(useCertKey.not());

        if (os != null && os == OSDetector.OS.WIN) {
            //show a warning when using multiple keys on windows, for PuTTY users
            warnMultiKeyLabel.visibleProperty().bind(useCustomKey.and(useCertKey));
        } else {
            warnMultiKeyLabel.setVisible(false);
        }

        pubKeyInPrivKeyInfo.visibleProperty().bind(state.isEqualTo(SshAuthState.PRIV_OPENSSH_PUB_INPRIV));

        BooleanBinding requirePubKeyField = state.isEqualTo(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_AUTO)
                .or(state.isEqualTo(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_MANUAL));
        pubKeyFileTextField.visibleProperty().bind(requirePubKeyField);
        pubKeyBrowseButton.visibleProperty().bind(requirePubKeyField);
        pubKeyLabel.visibleProperty().bind(requirePubKeyField);

        BooleanBinding requireDisabledPubKeyField = state.isEqualTo(SshAuthState.PRIV_OPENSSH_PUB_EXTERNAL_MANUAL).not();
        pubKeyFileTextField.disableProperty().bind(requireDisabledPubKeyField);
        pubKeyBrowseButton.disableProperty().bind(requireDisabledPubKeyField);
        pubKeyLabel.disableProperty().bind(requireDisabledPubKeyField);

        bindManagedToVisible(warnNoLoginKeyLabel);
        bindManagedToVisible(warnMultiKeyLabel);
        bindManagedToVisible(warnNoLoginKeyLabel);
        bindManagedToVisible(pubKeyBrowseButton);
        bindManagedToVisible(pubKeyLabel);
        bindManagedToVisible(pubKeyFileTextField);
        bindManagedToVisible(pubKeyInPrivKeyInfo);
    }

    @Override
    public Node getRoot() {
        return root;
    }

    private void setFileChooserDir(@Nonnull FileChooser fc, File... possibilitiesArr) {
        try {
            for (File file: possibilitiesArr) {
                if (file == null) {
                    continue;
                }
                if (file.exists() && file.isDirectory()) {
                    fc.setInitialDirectory(file);
                    return;
                }
                File parent = file.getParentFile();
                if (parent.exists() && parent.isDirectory()) {
                    fc.setInitialDirectory(parent);
                    return;
                }
            }
        } catch (Exception e) {
            LOG.error("Ignoring exception setting filechooser dir", e);
        }
    }

    @FXML
    protected void onBrowsePrivateKey() {
        FileChooser fc = new FileChooser();
        fc.setTitle("Select your Private Key");
        setFileChooserDir(fc,
                privateKeyFilename.get() == null ? null : new File(privateKeyFilename.get()),
                getLoggedInUserPrivateKeyFile(),
                new File(System.getProperty("user.home"))
        );

        File selectedFile = fc.showOpenDialog(privKeyFileTextField.getScene().getWindow());
        if (selectedFile != null) {
            loadPrivKey(new File(selectedFile.getAbsolutePath()));
            updateStatus();
            if (badPrivateKey.get()) {
                JFDialogs.create()
                        .owner(root)
                        .message(" Please select a valid private key.")
                        .masthead("The provided private key is not valid.")
                        .title("Invalid private key file")
                        .showWarning();
            }
        }
    }

    @FXML
    protected void onBrowsePublicKey() {
        FileChooser fc = new FileChooser();
        fc.setTitle("Select your Public Key");
        setFileChooserDir(fc,
                publicKeyFilename.get() == null ? null : new File(publicKeyFilename.get()),
                privateKeyFilename.get() == null ? null : new File(privateKeyFilename.get()),
                getLoggedInUserPrivateKeyFile(),
                new File(System.getProperty("user.home"))
        );

        File selectedFile = fc.showOpenDialog(pubKeyFileTextField.getScene().getWindow());
        if (selectedFile != null) {
            loadPubKey(new File(selectedFile.getAbsolutePath()));
            updateStatus();
            if (badPublicKey.get()) {
                JFDialogs.create()
                        .owner(root)
                        .message(" Please select a valid public key.")
                        .masthead("The provided public key is not valid.")
                        .title("Invalid public key file")
                        .showWarning();
            }
        }

    }

    protected File getLoggedInUserPrivateKeyFile() {
        return geniUserProvider.getLoggedInGeniUser().getPrivateKeyFile();
    }

    public boolean check() {
        if (!useCustomKey.get() && !useCertKey.get()) {
            JFDialogs.create()
                    .owner(root)
                    .message("You need to provide at least one key")
                    .showError();
            return false;
        }

        if (useCustomKey.get()) {
            if (badPrivateKey.get()) {
                JFDialogs.create()
                        .owner(root)
                        .message("The provided private key file is not valid")
                        .showError();
                return false;
            }

            if (requirePublicKeyfile.get() && badPublicKey.get()) {
                JFDialogs.create()
                        .owner(root)
                        .message("The provided public key file is not valid")
                        .showError();
                return false;
            }
        }

        return true;
    }

    @Override
    public boolean save() {
        if (!check()) {
            return false;
        }

        jFedPreferences.setBoolean(CorePreferenceKey.PREF_SSH_USE_LOGIN_KEY, useCertKey.get());

        jFedPreferences.setBoolean(CorePreferenceKey.PREF_SSH_USE_CUSTOM_KEY, useCustomKey.get());
        if (useCustomKey.get()) {
            File privKeyFile = new File(privateKeyFilename.get());
            assert privKeyFile.exists() && privKeyFile.isFile();
            if (state.get() == SshAuthState.PRIV_PUTTY) {
                jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PRIVATE_KEY_FILE);
                jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PUBLIC_KEY_FILE);
                jFedPreferences.setFile(CorePreferenceKey.PREF_SSH_PUTTY_KEY_FILE, privKeyFile);
            } else {
                jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PUTTY_KEY_FILE);
                jFedPreferences.setFile(CorePreferenceKey.PREF_SSH_PRIVATE_KEY_FILE, privKeyFile);
                if (requirePublicKeyfile.get()) {
                    jFedPreferences.setFile(CorePreferenceKey.PREF_SSH_PUBLIC_KEY_FILE, new File(publicKeyFilename.get()));
                } else {
                    jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PUBLIC_KEY_FILE);
                }
            }
        } else {
            jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PUTTY_KEY_FILE);
            jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PUBLIC_KEY_FILE);
            jFedPreferences.clear(CorePreferenceKey.PREF_SSH_PRIVATE_KEY_FILE);
        }

        //Ok, finally, register keys with pageant
        if (os != null && os == OSDetector.OS.WIN &&
                jFedPreferences.getBoolean(HLPreferenceKey.PREF_SSHAGENT_USE) == Boolean.TRUE) {
            Set<PuTTYFilesKeyInfo> keys = new HashSet<>();

            assert geniUserProvider.isUserLoggedIn();

            PuTTYFilesKeyInfo sshKeyInfo = jFedPreferences.getCustomSshKeyInfo();
            if (sshKeyInfo == null) {
                final GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
                sshKeyInfo = SshKeyInfoFactory.createGeniUserSshKeyInfo(geniUser);
            }
            keys.add(sshKeyInfo);

            if (!jFedPreferences.getString(CorePreferenceKey.PREF_SSHPROXY_USE_FOR_SSH, "UNKNOWN")
                    .equalsIgnoreCase("NEVER")) {
                final GeniUser geniUser = geniUserProvider.getLoggedInGeniUser();
                keys.add(SshKeyInfoFactory.createGeniUserSshKeyInfo(geniUser));
            }

            pageantHelper.registerKeys(keys, sshKeyInfoUnlocker);
        }

        return true;
    }

    protected boolean testOpenSshPrivateKeyFile(@Nonnull File file, @Nonnull String content) throws IOException {
        return KeyUtil.hasAnyPrivateKey(content);
    }

    protected boolean testPuttyPrivateKeyFile(@Nonnull File file, @Nonnull String content) throws IOException {
        return PuTTYPrivateKeyFile.isPuttyPrivateKey(content);
    }

    protected boolean testPrivateKeyFile(@Nonnull File file, @Nonnull String content) throws IOException {
        assert file.exists() && file.isFile();
        assert !content.trim().isEmpty();
        return testPuttyPrivateKeyFile(file, content) || testOpenSshPrivateKeyFile(file, content);
    }

    protected String getPublicKeyFromFile(@Nonnull File file, @Nonnull String content) throws IOException, KeyUtil.PEMDecodingException {
        if (testPuttyPrivateKeyFile(file, content)) {
            return getPublicKeyFromPuTTYFile(file, content);
        }
        if (testOpenSshPrivateKeyFile(file, content)) {
            return getPublicKeyFromOpenSSHPrivKey(file, content);
        }
        return null;
    }

    protected boolean testOpenSSHPrivateKeyIncludesPublicKey(@Nonnull File file, @Nonnull String content) {
        //TODO check if OpenSSH format keys have public key details
        boolean isFormatWithPublicKeyDetails = KeyUtil.hasRsaPrivateKey(content);
        boolean isEncrypted = KeyUtil.hasEncryptedRsaPrivateKey(content);

        return isFormatWithPublicKeyDetails && !isEncrypted;
    }

    protected String getPublicKeyFromOpenSSHPrivKey(@Nonnull File file, @Nonnull String content) throws KeyUtil.PEMDecodingException {
        RSAPrivateKey privKey = KeyUtil.pemToRsaPrivateKey(content, null);
        RSAPublicKey pubKey = KeyUtil.rsaPrivateKeyToRsaPublicKey(privKey);
        return KeyUtil.rsaPublicKeyToOpenSshAuthorizedKeysFormat(pubKey);
    }

    protected String getPublicKeyFromPuTTYFile(@Nonnull File file, @Nonnull String content) {
        String key = null;

        //try to extract public key from this file
        try {
            if (PuTTYPrivateKeyFile.isPuttyPrivateKey(file)) {
                PuTTYPrivateKeyFile ppk = PuTTYPrivateKeyFile.read(file);
                key = ppk.getPublicKeyOpenSshString();
            } else if (KeyUtil.isOpenSshRsaKey(file)) {
                key = IOUtils.fileToString(file).replace("\n", "");
            }
        } catch (IOException ex) {
            JFDialogs.create()
                    .owner(privKeyFileTextField)
                    .message("An error occured while processing the key file")
                    .masthead("Error")
                    .title("Error")
                    .showException(ex);

        }
        return key;
    }
}
